from schema import Schema, Optional, Or, And
from manifest_edit.context import Context
import re


class ManifestIterator:

    _MATCHALL_REGEX = r'.*'
    _KEY_WILDCARD = '*'
    _NEGATIVE_SELECTION_STRING = r''

    _recurse_keys = ['periods', 'adaptationSets', 'representations']

    def _schema(self):

        return Schema(
            {
                # 'periods' is mandatory. After this, you either have the
                #   specific plugin configuration or (possibly multiple)
                #   couples of <key> : <regular expression> couples, followed
                #   by the 'adaptationSets' keyword.
                'periods': Or(self._plugin_config_schema, [
                    And({
                        str: lambda s: re.compile(s),
                        Optional('adaptationSets'): Or(self._plugin_config_schema, [
                            And({
                                str: lambda s: re.compile(s),
                                Optional('representations'): Or(
                                    self._plugin_config_schema,
                                    [And({str: lambda s: re.compile(s)},
                                         lambda x: len(x.keys()) > 1
                                         )
                                     ]
                                )
                            }, lambda x: len(x.keys()) > 1)
                        ])
                    }, lambda x: len(x.keys()) > 1
                    )
                ])
            }
        )

    def __init__(self, configuration, plugin_config_schema=dict):
        '''
        This constructor can raise. It must be checked.
        '''
        self._plugin_config_schema = plugin_config_schema
        self.config = self._schema().validate(configuration)
        self._is_valid = self._schema().is_valid(self.config)

    def is_valid(self):
        return self._is_valid

    def __call__(self, root_node, process_keys, config=None):
        '''
        This method will expect a nested configuration dictionary that the
        user would have configured to express which particular list or lists
        (in the many available in the libfmp4 manifest_t type), he wants to
        process, and how. [config argument]

        This method will expect also a list of "process_keys", which
        the developer will have chosen to express the parameters of the
        particular manifest manipulation he needs to perform on the element
        or on the list. These will be used as a "sentinel" to stop recursion
        on the config dictionary and yield the corresponding manifest item or
        list.

        Depending on how the configuration file is written, this method
        will yield entire lists or single elements. It is responsibility
        of the developer to expect the right return type based on the
        cfg file he wrote.

        The expected structure is that the configuration
        dictionary contains a key nesting which is:

        - periods (level 0) OR adaptationSets (level 1) OR representations
        (level 2)

        Notice that level 0 is the only mandatory one.

        The corresponding value would be either of:

        - a list of dictionaries, where each item of the list will contain a
        selection key (i.e. id : "1", id : "*") and either of:
            - the process_key dictionary:
                in this case individuals item satisfying the selection will be
                yield
            - one of the possible recurse keys [periods, adaptationSets,
            representations]
                in this case the entire processing will restart from the next
                level
        - the process_key dictionary:
            in this case the entire list will be yield

        Example 1 of a valid config:
        periods:
            - id : "*"
            adaptationSets:
                - contentType :  "video"
                representations:
                    orderBy :      "bandwidth"
                    orderCriteria: "desc"
                - contentType :  "audio"
                representations:
                    orderBy :      "bandwidth"
                    orderCriteria: "desc"

        that would be passed to this method as:

            manifestIterator(
                manifest, config, ['orderBy', 'orderCriteria']
                )

        The calling method would receive "representations" lists, taken from
        "any" periods and from adaptationSets having type "video" or "audio",
        along with the {'orderBy': 'bandwidth, 'orderCriteria': 'desc'}
        dictionary.

        Example 2 of a valid config:
        periods:
            - id : "*"
            adaptationSets:
                - contentType :  "video"
                do :      "this"
                - contentType :  "audio"
                do :      "that"

        that would be passed to this method as:

            manifestIterator(
                manifest, config, ['do']
                )

        The calling method would receive individual adaptatation_set_t objects
        (not the entire "adaptationSets" lists), taken from
        "any" periods and having contentType "video" or "audio". Each of those
        will be yielded with the corresponding {'do' :'this'} or {'do' :'that'}
        dictionary.
        '''
        if not self._is_valid:
            yield None, None
            return

        if not config:
            config = self.config

        for node_name, node_config in config.items():

            # This is periods, adaptationSets or representations
            node = getattr(root_node, node_name)

            # This happens when you are in a "leaf" of a configuration that
            #  will return an entire list, not each individual elements.
            # That is, something like this:
            #
            #      periods:
            #       do :      "this"
            #
            # is going to yield the entire manifest.periods list
            if isinstance(node_config, dict):
                if all(key in process_keys for key in node_config.keys()):
                    # No selection key, just yield everything in list form
                    # Covered by test_period_selection_wildcard and many
                    #  others
                    yield node_config, node
                else:
                    # This can only happen if you are trying to find
                    #  configuration keys that are not present in the config
                    #  file. Could happen.
                    Context.log_debug(
                        'The provided process keys {p_ks} \
                       do not match config keys {c_ks}'.format(
                            p_ks=process_keys,
                            c_ks=node_config.keys()
                        ))

            # When instead you have a configuration that has some sort of
            #  selection items, individual elements will be returned
            # That is, something like this:
            #
            #      periods:
            #      - '*' : '*'
            #        do :      "this"
            #
            # is going to yield each individual period from the
            # manifest.periods list
            elif isinstance(node_config, list):
                for item in node_config:
                    # This is a dictionary, with multiple keys.
                    # Some are the processing keys (the plugin cfg keys) that
                    # is the variable process_keys;
                    # one can be a recurse_key;
                    # others are fields on which to perform item selection

                    # Keys that can be used to perform selection on this node.
                    item_selection_keys = [
                        key for key in item.keys()
                        if key not in self._recurse_keys+process_keys
                    ]

                    # If this node will be selected, you will have to recurse
                    #  these elements too
                    node_recurse_keys = [
                        key for key in item.keys()
                        if key in self._recurse_keys
                    ]
                    # unless you are in a leaf
                    node_recurse_key = node_recurse_keys[0] if len(
                        node_recurse_keys) else None
                    # Iterate over the manifest node (periods, adaptationSets
                    # or representations) and look for items with the
                    # configured property and value
                    for node_item in node:
                        # Notice that if the configuration specifies a property
                        # that does not exist in the manifest item, then this
                        # will not be selected, even if its configured value
                        # is '*'.
                        # If instead the property name is '*', then it will
                        # always be selected.
                        if self.node_matches(item_selection_keys,
                                             node_item,
                                             item
                                             ):
                            if node_recurse_key:
                                # Covered by test_period_selection_by_value and
                                # test_adaptation_set_selection_by_value
                                # Covered by
                                # test_adaptation_set_selection_wildcard
                                yield from self(
                                    node_item,
                                    process_keys,
                                    {node_recurse_key: item[node_recurse_key]}
                                )
                            else:
                                # Covered by test_selection_by_value_on_leaf
                                # Covered by
                                # test_adaptation_set_selection_wildcard_on_leaf
                                yield {k: item[k] for k in process_keys}, node_item
            # else:
                # No other checks on node_config are necessary if config
                #  validation works correctly

    def node_matches(self, item_selection_keys, node_item, item):
        # This function will return True if this manifest node_item matches
        #  the key, attribute pairs present in the the configuration "item"
        #
        matches_list = [self.node_matches_one(
            key, node_item, item) for key in item_selection_keys]
        return all(matches_list)

    def node_matches_one(self, item_selection_key, node_item, item):
        # We have a match if:
        # - the configured key is the wildcard value '*'
        # - the configured key exists in the manifest and its value is
        #   either matching the one specified in the config or the
        #   latter is '*'

        # Notice that this piece of code hides a very strong assumption, that
        #  seems to be used everywhere in libfmp4: any optional field
        #  is always present in manifest_t but has a default "illegal" value
        #  that is:
        #  - 0 for numeric fields
        #  - empty strings for string fields
        # We are going to leverage the fact that these values all evaluate to
        #  false-ish in python and just test for "node_value"

        # Quickly return in case of wildcard on key
        if item_selection_key == self._KEY_WILDCARD:
            return True

        # Let's see what is the value of this field
        node_value = getattr(
            node_item, item_selection_key, None)

        # And let's see what is the string specified by the user
        # You need str() conversion to allow users to just write 1 instead of
        # '1' in the config file
        # value_read_from_config_file = str(item[item_selection_key])
        value_read_from_config_file = item[item_selection_key]

        if not node_value:
            # If node_vaue evaluates to False-ish then it means this is an
            #  optional field and no value is present for it.
            # If the user has specified the 'empty string' value, it means
            #  he wants "negative selection": give me all those elements
            #  without a given field, so we have to return True
            if value_read_from_config_file == self._NEGATIVE_SELECTION_STRING:
                return True
            else:
                # Whatever other expression has been put by the user, if the
                #  field does not exist, we will return False
                return False

        # We know now that node_value is not False-ish, that is the manifest
        #  has a valid value for this field

        # If the users wanted all elements without this field, we must
        #  return False
        if value_read_from_config_file == self._NEGATIVE_SELECTION_STRING:
            return False

        # If the wildcard regex was selected, we can just return True
        # This is just an optimization assuming an if is faster than
        #  applying the regular expression
        if value_read_from_config_file == self._MATCHALL_REGEX:
            return True

        # Else we need to apply the regular expression to see if it
        #  matches
        # TODO optimize this so the regular expression is compiled once
        #  at configuration loading time
        return re.search(value_read_from_config_file, str(node_value))
